
/*
 *                            COPYRIGHT
 *
 *  PCB, interactive printed circuit board design
 *  Copyright (C) 1994,1995,1996 Thomas Nau
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Contact addresses for paper mail and Email:
 *  Thomas Nau, Schlehenweg 15, 88471 Baustetten, Germany
 *  Thomas.Nau@rz.uni-ulm.de
 *
 */


/* ---------------------------------------------------------------------------
 * This is the Notify Action that is included in the compilation of action.c
 *
 * does what's appropriate for the current mode setting. This normaly
 * means creation of an object at the current crosshair location.
 *
 * new created objects are added to the create undo list of course
 */
static void NotifyMode(void)
{
	void	*ptr1,
			*ptr2,
			*ptr3;
	int		type;

	if (Settings.RatWarn)
		ClearWarnings();
	switch(Settings.Mode)
	{
		case ARROW_MODE:
		{
			int	test;

			Note.Click = True;
				/* do something after click time */
			XtAppAddTimeOut(Context, CLICK_TIME, CB_Click, NULL);
				/* see if we clicked on something already selected
				 * (Note.Moving) or clicked on a MOVE_TYPE
				 * (Note.Hit)
				 */
			for (test = (SELECT_TYPES | MOVE_TYPES) & ~RATLINE_TYPE;
				test; test &= ~type )
			{
				type = SearchScreen(test, &ptr1, &ptr2, &ptr3);
				if (!Note.Hit && (type & MOVE_TYPES))
				{
					Note.Hit = type;
					Note.ptr1 = ptr1;
					Note.ptr2 = ptr2;
					Note.ptr3 = ptr3;
				}
				if (!Note.Moving && (type & SELECT_TYPES) &&
				TEST_FLAG(SELECTEDFLAG, (PinTypePtr) ptr2))
					Note.Moving = True;
				if ((Note.Moving && Note.Hit) || type == NO_TYPE)
					break;
			}
			break;
		}
			
		case VIA_MODE:
		{
			PinTypePtr	via;

			if (!PCB->ViaOn)
			{
				Message("You must turn via visibility on before\n"
					"you can place vias\n");
				break;
			}
			if ((via = CreateNewVia(PCB->Data, Crosshair.X, Crosshair.Y,
					Settings.ViaThickness, 2*GROUNDPLANEFRAME,
					Settings.ViaDrillingHole,
					NULL, VIAFLAG)) != NULL)
			{
				UpdatePIPFlags(via, (ElementTypePtr) via, NULL, NULL, False);
				AddObjectToCreateUndoList(VIA_TYPE, via, via, via);
				IncrementUndoSerialNumber();
				DrawVia(via, 0);
				Draw();
			}
			break;
		}

		case ARC_MODE:
		{
			switch(Crosshair.AttachedBox.State)
			{
				case STATE_FIRST:
					Crosshair.AttachedBox.Point1.X =
						Crosshair.AttachedBox.Point2.X = Crosshair.X;
					Crosshair.AttachedBox.Point1.Y =
						Crosshair.AttachedBox.Point2.Y = Crosshair.Y;
					Crosshair.AttachedBox.State = STATE_SECOND;
					break;

				case STATE_SECOND:
				case STATE_THIRD:
				{
					ArcTypePtr arc;
					int wx, wy, sa, dir;

					wx = Crosshair.X - Crosshair.AttachedBox.Point1.X;
					wy = Crosshair.Y - Crosshair.AttachedBox.Point1.Y;
					if (abs(wy) > abs(wx))
					{
						Crosshair.AttachedBox.Point2.X = Crosshair.AttachedBox.Point1.X
							+ abs(wy)*SGN(wx);
						sa = (wx >= 0) ? 0 : 180;
#ifdef ARC45
						if (abs(wy) >= 2*abs(wx))
							dir = (SGN(wx) == SGN(wy)) ? 45 : -45;
						else
#endif
							dir = (SGN(wx) == SGN(wy)) ? 90 : -90;
					}
					else
					{
						Crosshair.AttachedBox.Point2.Y = Crosshair.AttachedBox.Point1.Y
							+ abs(wx)*SGN(wy);
						sa = (wy >= 0) ? -90 : 90;
#ifdef ARC45
						if (abs(wx) >= 2*abs(wy))
							dir = (SGN(wx) == SGN(wy)) ? -45 : 45;
						else
#endif
							dir = (SGN(wx) == SGN(wy)) ? -90 : 90;
						wy = wx;
					}
					if (abs(wy) > 0 && (arc = CreateNewArcOnLayer(CURRENT,
						Crosshair.AttachedBox.Point2.X,
						Crosshair.AttachedBox.Point2.Y, abs(wy), sa, dir,
						Settings.LineThickness, 2*GROUNDPLANEFRAME, CLEARLINEFLAG)))
					{
						BoxTypePtr	bx;

						bx = GetArcEnds(arc);
						Crosshair.AttachedBox.Point1.X = 
						Crosshair.AttachedBox.Point2.X = bx->X2;
						Crosshair.AttachedBox.Point1.Y =
						Crosshair.AttachedBox.Point2.Y = bx->Y2;
						AddObjectToCreateUndoList(ARC_TYPE, CURRENT, arc, arc);
						IncrementUndoSerialNumber();
						addedLines++;
						DrawArc(CURRENT, arc, 0);
						Draw();
						Crosshair.AttachedBox.State = STATE_THIRD;
					}
					break;
				}
			}
			break;
		}
		case THERMAL_MODE:
		{
			int	LayerThermFlag;
			int	LayerPIPFlag = L0PIPFLAG << INDEXOFCURRENT;

			if (((type = SearchScreen(PIN_TYPES,&ptr1,&ptr2,&ptr3)) != NO_TYPE) &&
			    TEST_FLAG(LayerPIPFlag, (PinTypePtr) ptr3) &&
			    !TEST_FLAG(HOLEFLAG, (PinTypePtr) ptr3))
			{
				AddObjectToFlagUndoList(type, ptr1, ptr2, ptr3);
				LayerThermFlag = L0THERMFLAG << INDEXOFCURRENT;
				TOGGLE_FLAG(LayerThermFlag,(PinTypePtr) ptr3);
				IncrementUndoSerialNumber();
				ClearPin((PinTypePtr) ptr3, type, 0);
				Draw();
			}
			break;
		}

		case LINE_MODE:
				/* do update of position */
			NotifyLine();

				/* create line if both ends are determined && length != 0 */
			if (Crosshair.AttachedLine.State == STATE_THIRD)
			{
				LineTypePtr	line;

				if ((Crosshair.AttachedLine.Point1.X != Crosshair.AttachedLine.Point2.X
				    || Crosshair.AttachedLine.Point1.Y != Crosshair.AttachedLine.Point2.Y)
				    && (line = CreateNewLineOnLayer(CURRENT,
						Crosshair.AttachedLine.Point1.X,
						Crosshair.AttachedLine.Point1.Y,
						Crosshair.AttachedLine.Point2.X,
						Crosshair.AttachedLine.Point2.Y,
						Settings.LineThickness, 2*GROUNDPLANEFRAME,
						CLEARLINEFLAG)) != NULL)
				{
					PinTypePtr	via;

					addedLines++;
					AddObjectToCreateUndoList(LINE_TYPE, CURRENT, line, line);
					DrawLine(CURRENT, line, 0);
						/* place a via if vias are visible, the layer is
						   in a new group since the last line and there
						   isn't a pin already here */
					if (PCB->ViaOn && GetLayerGroupNumberByPointer(CURRENT) !=
						GetLayerGroupNumberByPointer(lastLayer) &&
						SearchObjectByPosition(PIN_TYPES, &ptr1, &ptr2, &ptr3,
							Crosshair.AttachedLine.Point1.X,
							Crosshair.AttachedLine.Point1.Y,
							Settings.ViaThickness/2) == NO_TYPE &&  
						(via = CreateNewVia(PCB->Data, Crosshair.AttachedLine.Point1.X,
						Crosshair.AttachedLine.Point1.Y, Settings.ViaThickness,
						2*GROUNDPLANEFRAME, Settings.ViaDrillingHole, NULL, VIAFLAG)) != NULL)
					{
						UpdatePIPFlags(via, (ElementTypePtr) via, NULL, NULL, False);
						AddObjectToCreateUndoList(VIA_TYPE, via, via, via);
						DrawVia(via, 0);
					}
						/* copy the coordinates */
					Crosshair.AttachedLine.Point1.X=Crosshair.AttachedLine.Point2.X;
					Crosshair.AttachedLine.Point1.Y=Crosshair.AttachedLine.Point2.Y;
					IncrementUndoSerialNumber();
					lastLayer = CURRENT;
				}
				if (PCB->Clipping && (Crosshair.X != Crosshair.AttachedLine.Point2.X
					|| Crosshair.Y != Crosshair.AttachedLine.Point2.Y) &&
					(line = CreateNewLineOnLayer(CURRENT,
						Crosshair.AttachedLine.Point2.X,
						Crosshair.AttachedLine.Point2.Y,
						Crosshair.X, Crosshair.Y,
						Settings.LineThickness, 2*GROUNDPLANEFRAME,
						CLEARLINEFLAG)) != NULL)
				{
					addedLines++;
					AddObjectToCreateUndoList(LINE_TYPE, CURRENT, line, line);
					IncrementUndoSerialNumber();
					DrawLine(CURRENT, line, 0);
						/* move to new start point */
					Crosshair.AttachedLine.Point1.X = Crosshair.X;
					Crosshair.AttachedLine.Point1.Y = Crosshair.Y;
					Crosshair.AttachedLine.Point2.X = Crosshair.X;
					Crosshair.AttachedLine.Point2.Y = Crosshair.Y;
					if (TEST_FLAG(SWAPSTARTDIRFLAG, PCB))
					{
						PCB->Clipping ^= 3;
						SetStatusLine();
					}
				}
				Draw();
			}
			break;

		case RECTANGLE_MODE:
				/* do update of position */
			NotifyBlock();

				/* create rectangle if both corners are determined 
				 * and width, height are != 0
				 */
			if (Crosshair.AttachedBox.State == STATE_THIRD &&
				Crosshair.AttachedBox.Point1.X != Crosshair.AttachedBox.Point2.X &&
				Crosshair.AttachedBox.Point1.Y != Crosshair.AttachedBox.Point2.Y)
			{
				PolygonTypePtr	polygon;

				if ((polygon = CreateNewPolygonFromRectangle(CURRENT, 
						Crosshair.AttachedBox.Point1.X,
						Crosshair.AttachedBox.Point1.Y,
						Crosshair.AttachedBox.Point2.X,
						Crosshair.AttachedBox.Point2.Y,
						CLEARPOLYFLAG)) != NULL)
				{
					AddObjectToCreateUndoList(POLYGON_TYPE, CURRENT,
						polygon, polygon);
					UpdatePIPFlags(NULL, NULL, CURRENT, polygon, True);
					IncrementUndoSerialNumber();
					DrawPolygon(CURRENT, polygon, 0);
					Draw();
				}

					/* reset state to 'first corner' */
				Crosshair.AttachedBox.State = STATE_FIRST;
			}
			break;

		case TEXT_MODE:
		{
			char	*string;

			if ((string = GetUserInput("Enter text:", "")) != NULL)
			{
				TextTypePtr	text;
				int		flag = NOFLAG;

				if (GetLayerGroupNumberByNumber(INDEXOFCURRENT) ==
				    GetLayerGroupNumberByNumber(MAX_LAYER + SOLDER_LAYER))
					flag = ONSOLDERFLAG;
				if ((text = CreateNewText(CURRENT, &PCB->Font, Crosshair.X,
						Crosshair.Y, 0, Settings.TextScale,
						string, flag)) != NULL)
				{
					AddObjectToCreateUndoList(TEXT_TYPE, CURRENT, text, text);
					IncrementUndoSerialNumber();
					DrawText(CURRENT, text, 0);
					Draw();
				}

					/* free memory allocated by GetUserInput() */
				SaveFree(string);
			}
			break;
		}

		case POLYGON_MODE:
		{
			PointTypePtr	points = Crosshair.AttachedPolygon.Points;
			Cardinal		n = Crosshair.AttachedPolygon.PointN;
			
				/* do update of position; use the 'LINE_MODE' mechanism */
			NotifyLine();

				/* check if this is the last point of a polygon */
			if (n >= 3 &&
				points->X == Crosshair.AttachedLine.Point2.X &&
				points->Y == Crosshair.AttachedLine.Point2.Y)
			{
				CopyAttachedPolygonToLayer();
				Draw();
				break;
			}

				/* create new point if it's the first one or if it's
				 * different to the last one
				 */
			if (!n ||
				points[n-1].X != Crosshair.AttachedLine.Point2.X ||
				points[n-1].Y != Crosshair.AttachedLine.Point2.Y)
			{
				CreateNewPointInPolygon(&Crosshair.AttachedPolygon,
					Crosshair.AttachedLine.Point2.X,
					Crosshair.AttachedLine.Point2.Y);

					/* copy the coordinates */
				Crosshair.AttachedLine.Point1.X=Crosshair.AttachedLine.Point2.X;
				Crosshair.AttachedLine.Point1.Y=Crosshair.AttachedLine.Point2.Y;
			}
			break;
		}

		case PASTEBUFFER_MODE:	
		{
			Cursor oldCursor;

			oldCursor = SetOutputXCursor(XC_watch);
			if (CopyPastebufferToLayout(Crosshair.X, Crosshair.Y))
				SetChangedFlag(True);
			SetOutputXCursor(oldCursor);
			break;
		}

		case REMOVE_MODE:
			if ((type = SearchScreen(REMOVE_TYPES,&ptr1,&ptr2,&ptr3)) != NO_TYPE)
			{
				if (type == ELEMENT_TYPE)
				{
					RubberbandTypePtr	ptr;
					int			i;

					Crosshair.AttachedObject.RubberbandN = 0;
					LookupRatLines(type, ptr1, ptr2, ptr3);
					ptr = Crosshair.AttachedObject.Rubberband;
					for (i = 0; i < Crosshair.AttachedObject.RubberbandN; i++)
					{
						if (PCB->RatOn)
							EraseRat((RatTypePtr) ptr->Line);
						MoveObjectToRemoveUndoList(RATLINE_TYPE,
							ptr->Line, ptr->Line, ptr->Line);
						ptr++;
					}
				}
				RemoveObject(type, ptr1, ptr2, ptr3);
				SetChangedFlag(True);
			}
			break;

		case ROTATE_MODE:
			Crosshair.AttachedObject.RubberbandN = 0;
			if ((type = SearchScreen(ROTATE_TYPES,&ptr1,&ptr2,&ptr3)) != NO_TYPE)
			{
				if (TEST_FLAG(RUBBERBANDFLAG, PCB))
					LookupRubberbandLines(type,ptr1,ptr2,ptr3);
				if (type == ELEMENT_TYPE)
					LookupRatLines(type, ptr1, ptr2, ptr3);
				if (ShiftPressed())
					RotateObject(type,ptr1,ptr2,ptr3, Crosshair.X,
						Crosshair.Y, SWAP_IDENT ? 1 : 3);
				else
					RotateObject(type,ptr1,ptr2,ptr3, Crosshair.X,
						Crosshair.Y, SWAP_IDENT ? 3 : 1);
				SetChangedFlag(True);
			}
			break;

			/* both are almost the same */
		case COPY_MODE:
		case MOVE_MODE:
			switch(Crosshair.AttachedObject.State)
			{
					/* first notify, lookup object */
				case STATE_FIRST:
				{
					int	types = (Settings.Mode == COPY_MODE ) ?
									COPY_TYPES : MOVE_TYPES;

					Crosshair.AttachedObject.Type = SearchScreen(types,
						&Crosshair.AttachedObject.Ptr1,
						&Crosshair.AttachedObject.Ptr2,
						&Crosshair.AttachedObject.Ptr3);
					if (Crosshair.AttachedObject.Type != NO_TYPE)
						AttachForCopy(Crosshair.X, Crosshair.Y);
					break;
				}

					/* second notify, move or copy object */
				case STATE_SECOND:
					if (Settings.Mode == COPY_MODE)
						CopyObject(Crosshair.AttachedObject.Type,
							Crosshair.AttachedObject.Ptr1,
							Crosshair.AttachedObject.Ptr2,
							Crosshair.AttachedObject.Ptr3,
							Crosshair.X -Crosshair.AttachedObject.X,
							Crosshair.Y -Crosshair.AttachedObject.Y);
					else
						MoveObjectAndRubberband(Crosshair.AttachedObject.Type,
							Crosshair.AttachedObject.Ptr1,
							Crosshair.AttachedObject.Ptr2,
							Crosshair.AttachedObject.Ptr3,
							Crosshair.X -Crosshair.AttachedObject.X,
							Crosshair.Y -Crosshair.AttachedObject.Y);
					SetChangedFlag(True);

						/* reset identifiers */
					Crosshair.AttachedObject.Type = NO_TYPE;
					Crosshair.AttachedObject.State = STATE_FIRST;
					break;
			}
			break;

			/* insert a point into a polygon/line/... */
		case INSERTPOINT_MODE:
			switch(Crosshair.AttachedObject.State)
			{
					/* first notify, lookup object */
				case STATE_FIRST:
					Crosshair.AttachedObject.Type =
						SearchScreen(INSERT_TYPES,
						&Crosshair.AttachedObject.Ptr1,
						&Crosshair.AttachedObject.Ptr2,
						&Crosshair.AttachedObject.Ptr3);

					if (Crosshair.AttachedObject.Type != NO_TYPE)
					{
							/* get starting point of nearest segment */
						if (Crosshair.AttachedObject.Type == POLYGON_TYPE)
						{
                                                        fake.poly = (PolygonTypePtr)
                                                                       Crosshair.AttachedObject.Ptr2;
                                                        polyIndex = GetLowestDistancePolygonPoint(
                                                                       fake.poly, Crosshair.X, Crosshair.Y);
                                                        fake.line.Point1 = fake.poly->Points[polyIndex];
                                                        fake.line.Point2 = (polyIndex) ?
                                                                          fake.poly->Points[polyIndex-1]
                                                                          : fake.poly->Points[fake.poly->PointN-1];
                                                        Crosshair.AttachedObject.Ptr2 = &fake.line;

						}
						Crosshair.AttachedObject.State = STATE_SECOND;
                                                AdjustInsertPoint();
					}
					break;

					/* second notify, insert new point into object */
				case STATE_SECOND:
                                        if (Crosshair.AttachedObject.Type == POLYGON_TYPE)
                                                InsertPointIntoObject(POLYGON_TYPE,
                                                        Crosshair.AttachedObject.Ptr1, fake.poly,
                                                        &polyIndex,
                                                        InsertedPoint.X, InsertedPoint.Y, False);
                                        else    
					InsertPointIntoObject(Crosshair.AttachedObject.Type,
						Crosshair.AttachedObject.Ptr1,
						Crosshair.AttachedObject.Ptr2,
						&polyIndex,
                                                InsertedPoint.X, InsertedPoint.Y, False);
					SetChangedFlag(True);

						/* reset identifiers */
					Crosshair.AttachedObject.Type = NO_TYPE;
					Crosshair.AttachedObject.State = STATE_FIRST;
					break;
			}
			break;
	}
}
